#define std_io_OutputTube_IMPORT

#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <err.h>
#include <std/io/impl/OutputTube.h>

#ifdef linux
/*
  Linuxˤfunopen(3)ʤΤǡfopencookie(3)ǥߥ졼Ȥ롣
*/
#include <libio.h>

typedef struct {
    void *cookie;
    int (*readfn)(void *, char *, int);
    int (*writefn)(void *, const char *, int);
    /* seekfnηˤĤƤfunopen()ΥȤ򻲾ȤΤ */
    off_t (*seekfn)(void *, off_t, int);
    int (*closefn)(void *);
} BSD;

static int 
GNU_read(void *gnu, char *data, size_t size)
{
    BSD *bsd = (BSD *)gnu;

    return bsd->readfn(bsd->cookie, data, (int)size);
}

static int 
GNU_write(void *gnu, const char *data, size_t size)
{
    BSD *bsd = (BSD *)gnu;

    return bsd->writefn(bsd->cookie, data, (int)size);
}

static int 
GNU_seek(void *gnu, _IO_off64_t *pos, int whence)
{
    BSD *bsd = (BSD *)gnu;

    return ((*pos = bsd->seekfn(bsd->cookie, *pos, whence)) == -1);
}

static int 
GNU_close(void *gnu)
{
    BSD *bsd = (BSD *)gnu;
    int s;

    s = bsd->closefn(bsd->cookie);
    free(bsd);
    return s;
}

static const cookie_io_functions_t GNU_FUNCTIONS = {
    .read = GNU_read,
    .write = GNU_write,
    .seek = GNU_seek,
    .close = GNU_close};

static FILE *
funopen(const void *cookie,
	int (*readfn)(void *, char *, int),
	int (*writefn)(void *, const char *, int),
	/*
	  ʪfunopen(3)4ϼΤ褦Ƥ롣

	  fpos_t (*seekfn)(void *, fpos_t, int),

	  BSDǤfpos_t64ӥåʤΤǡseekfnθƤӽФ
	  lseek(2)ȥޥåƤȹͤƤ褤GNU libcǤ
	  fpos_tϹ¤ΤǤ뤿ᡢ嵭Ǥseekfnlseek(2)θ
	  ӽФϥޥåʤlseek(2)2ͤηǤ
	  off_tfpos_tߤѴǤʤΤǡΤޤޤǤϥߥ
	  졼ǤʤȤˤʤ롣

	  Τᡢfunopen(2)4뤳Ȥˤ롣
	  衢νϥꥸʥAPIŬѤ٤Ǥ롣Ʊ
	  readfn, writefnθƤӽФ⤽줾read(2), write(2)
	  ˥ޥå2char *Ǥʤvoid *3int
	  Ǥʤsize_tˤʤˤ褦˽٤

	  Υ饤֥ǤseekfnϻѤʤΤǡνˤĤ
	  Ϥޤ꿼ͤɬפϤʤ¾ήѤȤդɬ
	  פ롣
	*/
	off_t (*seekfn)(void *, off_t, int),
	int (*closefn)(void *))
{
    BSD *bsd;

    if ((bsd = (BSD *)malloc(sizeof(*bsd))) == NULL)
	return NULL;
    bsd->cookie = (void *)cookie;
    bsd->readfn = readfn;
    bsd->writefn = writefn;
    bsd->seekfn = seekfn;
    bsd->closefn = closefn;
    return fopencookie(bsd, "r+", GNU_FUNCTIONS);
}
#endif /* linux */

void
OutputTube_initialize(union OutputTube out,
		      const struct OutputTube_Methods *methods)
{
    out.impl->methods = methods;
    out.impl->hasError = 0;
    out.impl->ancientStreams = 0;
}

ssize_t
OutputTube_write(union OutputTube out, const void *data, size_t size)
{
    ssize_t s;

    if ((s = out.impl->methods->write(out.impl, data, size)) < 0) {
	out.impl->hasError = 1;
	return -1;
    }
    return s;
}

int
OutputTube_writeFully(union OutputTube out, const void *top, size_t size)
{
    int s;
    const char *data = top;

    while (size > 0 && (s = OutputTube_write(out, data, size)) > 0) {
        data += s;
        size -= s;
    }
    return (size > 0) ? -1 : 0;
}

int
OutputTube_hasError(union OutputTube out)
{
    return out.impl->hasError;
}

int
OutputTube_close(union OutputTube out)
{
    if (out.impl->ancientStreams) {
	warnx("%s: %d ancient stream(s) not closed.", __FUNCTION__,
	      out.impl->ancientStreams);
    }
    return out.impl->methods->close(out.impl);
}

union AncientStream {
    void *cookie;
    union OutputTube out;
} __attribute__((__transparent_union__));

static int
writefn(union AncientStream this, const char *data, int size)
{
    return OutputTube_write(this.out, data, size);
}

static int
closefn(union AncientStream this)
{
    --(this.out.impl->ancientStreams);
    return 0;
}

FILE *
OutputTube_getAncientStream(union OutputTube out)
{
    FILE *fp;

    if ((fp = funopen(out.impl, NULL, writefn, NULL, closefn)) == NULL)
	return NULL;
    ++(out.impl->ancientStreams);
    return fp;
}

static int
print(union OutputTube out, const char *format, va_list ap)
{
    FILE *fp;
    int s;

    if ((fp = OutputTube_getAncientStream(out)) == NULL)
	return -1;
    s = vfprintf(fp, format, ap);
    return (fclose(fp) < 0 || s < 0) ? -1 : 0;
}

int
OutputTube_print(union OutputTube out, const char *format, ...)
{
    va_list ap;
    int s;

    va_start(ap, format);
    s = print(out, format, ap);
    va_end(ap);
    if (s < 0) {
	out.impl->hasError = 1;
	return -1;
    }
    return 0;
}
